#!/usr/bin/env python
# coding=utf-8

import logging

TIMESTMP_FMT = ("%H:%M:%S") # no %f for msecs
FORMAT = '%(asctime)s.%(msecs)03d %(name)-5s %(message)s'
logging.basicConfig(format=FORMAT, datefmt=TIMESTMP_FMT, level=logging.INFO)

log = logging.getLogger("Classes")
log.setLevel(logging.INFO)# DEBUG complains about unused args in __repr__ % self.__dict__

# forward declarations
class State:
    pass
class Sensor:
    pass
class TerminalBlock:
    pass
class Output:
    pass
class InputBinaryState:
    pass
class InputPushButton:
    pass

"-------- helper -------------"
class DSID(object):
    def __init__(self, _list):
        self.dsid = _list
    def __repr__(self):
        return ":".join(map(lambda x: "%02X" % x, self.dsid))
        # return list2str(self.dsid)

"----- structure classes ------"
"dass die gruppen entweder am apartment oder an der zone haengen, also nur eine \
 der verbindungen ist gueltig pro instanz"
class Apartment:
    apartmentId = "DS_UUID"

    " --- DSS ---"
    errorTemperatureLimit = ""
    warnTemperatureLimit = ""
    id_ = "probably short of apartmentId"
    isProgrammingModeLocked = bool()

    " --- has-a ---"
    user_groups = range(16, 24), "user created group"
    #apartmentScenes = [key for (key, name) in StandardScene if name in apartment_scene]
    states = [State(), State()]

class Zone(object):
    '''
    "zero means all zones"
    apartmentId = ""
    zoneId = ""
    name = ""

    " --- has-a ---"
    meters = list()
    zoneScenes = [key for (key, name) in StandardScene if name in zone_scene]
    states = [ State(), State() ]
    groups
    devices = [ DSDevice(), ] # TODO or TerminalBlock in 2.0
    '''

    def __init__(self, zoneid, name):
        self.zoneid = zoneid
        self.name = name
        self.meters = list()

    def addDsm(self, dsm):
        meters = self.meters
        if meters.count(dsm) == 0:
            meters.append(dsm)

class Group(object):
    '''
    name
    apartmentId = 0
    zoneId = 0
    groupId = 0 00..63
    userDefinedGroup: bool
    '''
    "groups are apartment wide, hence groupid itself is not unique"
    apartmentId = 0
    zoneId = 0
    groupId = 0

    def __init__(zoneId, groupId):
        self.zoneId = zoneId
        self.groupId = groupId

    def userDefinedGroup(self):
        return bool(groupId in range(64, 72)) # TODO contradicts groupId < 64

    " -- has-a --- "
    sensors = list()
    #groupScenes = [key for (key, name) in StandardScene if (name not in zone_scene and \
    #                                                   name not in apartment_scene) ]
    devices = list()
    group = list()


class FunctionGroup:
    '''
    A group is the functional class of electric devices. digitalSTROM defines
    basic groups to reflect the common usage of electric devices in buildings.
    Those predefined groups allow to have a simple system setup as all digital-
    STROM Ready Devices have their standard group preconfigured.

    abbreviation: char(3)
    color: string(10)
    description string(100)
    name string(10)
    functionGroupId  1-15"
    '''
    def __init__(self, functionGroupId, name, color, description):
        self.functionGroupId = functionGroupId
        self.name = name
        self.color = color
        self.description = description
        # TODO abbreviaton


class Scene(object):
    pass

class State(object):
    lastUpdate = ""
    name = ""
    value = "stateValue"
    pass

"------ hardware classes -------"

class DSServer:
    pass

class DSFilter:
    pass

class ConnectionState:
    not_connected = 'not connected'
    synchronizing = 'synchronizing'
    active = 'active'
    inactive = 'inactive' # what's the difference to not_connected
    incompatible = 'incompatible' # my invention

class DSMeter:
    '''
    apiVersion = ""
    gtin = "DsmApiGetDeviceDescription()"
    hardwareVersion = ""
    name = ""
    serialNumber = ""
    softwareArmVersion = ""
    softwareDSPVersion = ""
    state = ConnectionState

    " --- DSS --- "
    dataModelHash = "changes with every reboot"
    dataModificationCount = "changes with every device change"
    frequencyPair = ""
    inputEventCount = ""
    ledConfiguration = ""
    pingInterval = ""
    pingMode = ""

    dsid = "key"

    " --- has-a --- "
    belongsToApartment = Apartment()
    consumption = Sensor()
    zones = [ Zone(), Zone()]
    terminalBlocks = [ TerminalBlock(), TerminalBlock() ]
    '''
    gtin = "DsmApiGetDeviceDescription()"
    serialNumber = ""
    state = ConnectionState.active

    " --- DSS --- "
    dataModelHash = "changes with every reboot"
    dataModificationCount = "changes with every device change"
    frequencyPair = ""
    inputEventCount = ""
    ledConfiguration = ""
    pingInterval = ""
    pingMode = ""

    def __init__(self, record):
        self.updateVersion(record)

    def updateVersion(self, record):
        (self.hardwareVersion, self.softwareArmVersion,
         self.softwareDSPVersion, self.apiVersion, self.dsid, self.name) = record

    def updateConfigHash(self, dataModelHash, dataModificationCount):
        self.dataModelHash = dataModelHash
        self.dataModificationCount = dataModificationCount

    " --- has-a --- "
    belongsToApartment = Apartment()
    consumption = Sensor()
    #zones = [ Zone(), Zone()]
    terminalBlocks = [ TerminalBlock(), TerminalBlock() ]

    def __str__(self):
        return "%s %s " % (DSID(self.dsid).serial(), self.name)

    def __repr__(self):
        ret = str(self)
        vs = map(format_version, [self.hardwareVersion,
                                  self.softwareArmVersion,
                                  self.softwareDSPVersion, self.apiVersion])
        ret += "\n\tdsid: %s" % dsid_str(self.dsid)
        ret += "\n\thwv: %s  armv: %s  dspv: %s  apiv: %s " % tuple(vs)
        return ret

class InternetState:
    # device needs internet access?
    invalid = 0x0 # EAN not configured in EEPROM
    noInternetAccess = 0x1
    InternetOptional = 0x2
    InternetMandatory = 0x3  # der pankibutton mit securitas-service"

class OemState:
    fetching = "fetching"  # e.g. device inactive
    invalid = "no EAN configured"
    valid =  "EAN valid"


"das EAN feld ist in allen beteiligten terminal-blocks gleich; erst dadurch \
 werden die terminalblocks zu einem dsdevice"
class DSDevice:
    '''
    gtin:       ean feld in der klemme
    internetState "benoetigt das das dsdevice internet: noEAN, \
                   noInternetAccess, InternetOptional, InternetMandatory, \
                   der pankibutton mit securitas-service"
    serialNumber ""
    dsid        "key" / what if multi-tb device

    iconPath    ""
    productInfo ""
    name:       ""
    sgtin:      self.gtin + self.serial

    " -- cloud --"
    oemSerialNumber = 0

    "-- has-a --"
    terminalBlocks = list()
    isStandardGroup = Group(), "TODO why not the terminal block"
    belongsToGroups = [Group(), Group()]
    belongsToZone = Zone()

    " -- mine --"
    oemAreaState = [ fetching, invalid, valid ]
    '''

    "-- DSS --"
    iconPath = ""
    productInfo = ""
    name = ""

    def sgtin(self):
        return self.gtin + self.serialNumber

    def __init__(self, groups, terminalBlocks, oemAreaState =
                 OemState.fetching, oemEan = None, oemInetNeed = None,
                 oemSerialNumber = None, dsid = None, serialNumber = None):
        log.debug("DSDevice.__init__ %s %s %s %s %s %s %s %s, %s\n",  groups,
                 terminalBlocks, oemAreaState, OemState.fetching, oemEan,
                 oemInetNeed, oemSerialNumber, dsid, serialNumber)
        self.groups = groups
        self.oemAreaState = oemAreaState
        self.terminalBlocks = terminalBlocks
        if oemAreaState == OemState.valid:
            self.gtin = oemEan
            self.internetState = oemInetNeed
            self.oemSerialNumber = oemSerialNumber
            self.oemEan = oemEan # needed?
        else:
            # either fetching or invalid
            self.serialNumber = serialNumber # fallback to DSID
            self.dsid = dsid

    def updateOEM(self, oemEan, oemInetNeed, oemSerialNumber):
        log.info('DSDevice.updateOEM %s %s %s', oemEan, oemInetNeed, oemSerialNumber)
        self.oemAreaState = oemAreaState.valid
        self.gtin = oemEan
        self.internetState = oemInetNeed
        self.oemSerialNumber = oemSerialNumber
        self.oemEan = oemEan # needed?

    "-- has-a --"
    terminalBlocks = list()
    #isStandardGroup = Group(), "TODO why not the terminal block"
    #belongsToGroups = [Group(), Group()]
    #belongsToZone = Zone()

    def __repr__(self):
        ret = "DSDevice: "
        if self.oemAreaState == OemState.valid:
            ret += "gtin %(gtin)s %(internetState) %(oemSerialNumber)\n" % self.__dict__
        else:
            ret += "dsid %(dsid)s serial: %(serialNumber)d\n" % self.__dict__

        ret += "OEM Area : %(oemAreaState)s\n" % self.__dict__
        # TODO productInfo state

        ret += "\tgroups: [%s]\n" % ", ".join(map(str, self.groups))
        for i in self.terminalBlocks:
            ret += "\t" + i.__str__() + "\n"

        return ret


" gtin -- global trade item number"
# dsid (=SGTIN)
class TerminalBlock:
    '''
    firmwareRevision = ""
    firstSeen = ""
    gtin = "DsmApiGetDeviceDescription()"
    hardwareInfo = ""
    icon = "iconUrl"
    isIndependent --  zeigt an ob ein terminal-block unabhängig des dsdevice
                verwendet werden kann: z.b. ein integrierter taster
    lastDiscovered = "date"
    lastKnownDSM = "dsid"
    partNumber --  dsdevice.terminal_blocks[partNumber] == self
    room = "room12"
    serialNumber = 12
    "-- runtime --"
    dsm = "dsmid"
    shortAddress = 914
    state = "connectionState"

    "-- has-a --"
    consumption = Sensor(), "TODO, verify"
    #belongsToZone = Zone(), "might change with 2.0, might be physical room"
    partOfDevice = DSDevice([])
    output = Output(),  "maximally 1x"
    sensors = [Sensor()], "0..14"
    inputBinaryStates = [InputBinaryState(),], "0..16"
    inputPushButtons = [InputPushButton(), ], "0..4"

    -- mine --
    zoneId
    '''

    firstSeen = "" # TODO
    icon = "iconUrl"
    lastDiscovered = "date"
    lastKnownDSM = "dsid" # TODO
    room = "room12"

    def __init__(self, deviceId, firmwareRevision, zoneId, connectionState,
                 serialNumber, dsmid, gtin, hardwareInfo):
        self.firmwareRevision = firmwareRevision
        self.serialNumber = serialNumber
        self.dsmid = dsmid
        self.shortAddress = deviceId
        self.deviceId = self.shortAddress # TODO remove
        self.state = connectionState
        self.zoneId = zoneId
        self.gtin = gtin
        self.hardwareInfo = hardwareInfo
        self.inputBinaryStates = list()
        self.inputPushButtons = list()
        self.oemAreaState = OemState.fetching

    def updateOEM(self, oemAreaState, isIndependent=None, oemPartNumber=None):
        log.info('TerminalBlock.updateOEM %s %s %s', oemAreaState, isIndependent, oemPartNumber)
        self.oemAreaState = oemAreaState
        if oemAreaState == OemState.valid:
            self.isIndependent = isIndependent
            self.partNumber = oemPartNumber

    def dsid(self):
        return self.gtin + self.serial

    def setOutput(self, output):
        self.output = output

    def appendInputPushButton(self, pushButton):
        self.inputPushButtons.append(pushButton)

    def appendInputBinaryState(self, binaryInput):
        self.inputBinaryStates.extend(binaryInput)

    "-- has-a --"
    consumption = Sensor(), "TODO, verify"
    #belongsToZone = Zone(), "might change with 2.0, might be physical room"
    partOfDevice = DSDevice([], [], OemState.invalid, [53, 00], 0x1234)
    sensors = [Sensor()], "0..14"

    def __str__(self):
        return "t-block: %(shortAddress)d info: %(hardwareInfo)s state: %(state)s" % self.__dict__

    def __repr__(self):
        ret = \
        "t-block: %(shortAddress)d info: %(hardwareInfo)s state: %(state)s\n" + \
        "\tgtin %(gtin)s serial %(serialNumber)s version 0x%(firmwareRevision)x\n" + \
        "\tdsm: %(dsmid)s discovered: %(lastDiscovered)s\n" + \
        "\tlastKnownDSM %(lastKnownDSM)s firstSeen: %(firstSeen)s\n"

        cp = dict(self.__dict__)

        if self.oemAreaState == OemState.valid:
            ret += "\tOEM data: partNumber %(partNumber)d isIndependent %(isIndependent)d"
        else:
            ret += "\tOEM data: %(oemAreaState)s"

        # sneak in yet missing fields
        cp.update(dict(lastDiscovered = "", lastKnownDSM="", icon = "",
                       firstSeen = ""))
        ret = ret % cp


        if hasattr(self, 'output'):
            ret += "\n\t%s" % str(self.output)
        for i in self.inputPushButtons:
            ret += "\n\t%s" % str(i)
        for i in self.inputBinaryStates:
            ret += "\n\t%s" % str(i)

        return ret



class Sensor_:
    '''
    name = ""
    _type = [ 'current' ]
    '''
    name = ""
    _type = [ 'current' ]

    def __repr__(self):
        return "SENSOR"

class Output:
    '''
    mode = [ 'switched', 'dimmed', 'motor', 'wischen', 'jalousie', 'blinds', 'disabled' ]
    name = ""
    type_ = ['brightness', 'color', 'height', 'position', 'opacity']
    '''
    #group = Group()
    mode = [ 'switched', 'dimmed', 'motor', 'wischen', 'jalousie', 'blinds', 'disabled' ]
    name = ""
    type_ = ['brightness', 'color', 'height', 'position', 'opacity']

    def __init__(self, outputMode):
        self.outputMode = outputMode # todo convert properly

    def __repr__(self):
        return "OUTPUT mode: %(outputMode)s" % self.__dict__

" lt:= lokal taster, sagt welche szenen dieser taster ausloest"
" mode := e.g. dimming, switching"
class InputBinaryState:
    '''
    mode = "local, room, area, extended, apartment, app"
    name = ""
    sensorFunction = "aka inputType: smoke detector, windmonitor, rain detector, etc "
    targetGroup = ""
    targetType = ""

    "-- dS ---"
    ltMode = "lokal taster, TODO how can input state have local taster"
    '''

    # missing: mode, ltMode, name
    def __init__(self, targetType, targetGroup, inputType, inputId):
        self.targetType = targetType
        self.targetGroup = targetGroup
        self.sensorFunction = inputType
        # TODO self.ltMode = inputId

    def __repr__(self):
        return "BinaryState: type %s group %s" % (self.targetType, self.targetGroup)

" callsPresent -- implicit present state change"
" lt:= lokal taster, sagt welche szenen dieser taster ausloest"
" mode := e.g. dimming, switching"
class InputPushButton:
    '''
    callsPresent = "aka coming home button, somebody in the room"
    group = Group()
    mode = "local, room, area, extended, apartment, app"
    name = ""
    setsLocalPriority = ""

    " --- dS ---"
    ltMode = ""
    '''
    callsPresent = "aka coming home button, somebody in the room"
    #group = Group()
    mode = "local, room, area, extended, apartment, app"
    name = ""
    setsLocalPriority = ""

    def __init__(self, activeGroup, setLocalPriority, callsPresent):
        self.callsPresent = callsPresent
        self.activeGroup = activeGroup
        self.setsLocalPriority = setLocalPriority

    def __repr__(self):
        ret = "BUTTON group: %(activeGroup)d presenceDetector: %(callsPresent)s " + \
            "local priority: %(setsLocalPriority)s"
        return ret % self.__dict__

    " --- dS ---"
    ltMode = ""



" ---- Runtime classess ---- "

class State:
    lastUpdate = ""
    name = ""
    value = "stateValue"



" --- unit test ---- "

if __name__=="__main__":
    dsm = DSMeter((184620288L, 477337377L, 3151257746L, 519, [53, 4, 23, 95, 224, 0, 0, 16, 0, 0, 21, 10], 'dSM1'))
    print repr(dsm)

    print Output("foo")

    print InputPushButton(2, False, False)

    tb = TerminalBlock(913, 0x350, 0, ConnectionState.active, 0xffff,
                       [0xff, 0xff], 0x2222222222, "GE-TKM120")

    repr(tb)
    str(tb)
    tb.updateOEM(False, 0xffff)
    repr(tb)
    str(tb)
    tb.setOutput(Output("foo"))
    repr(tb)
    str(tb)
    tb.appendInputPushButton(InputPushButton(2, False, False))
    repr(tb)
    str(tb)
    tb.appendInputBinaryState([InputBinaryState("1", "2", "3", 4)])
    repr(tb)
    str(tb)

    devSpec = dict(
        groups = [1, 2, 3, 4],
        terminalBlocks = [tb],
        oemAreaState = OemState.fetching,
        dsid = [ 53, 26, 0, 0],
        serialNumber = 0x12345,
    )

    repr(DSDevice(**devSpec))

    devSpec = dict(
        groups = [1, 2, 3, 4],
        terminalBlocks = [tb],
        oemAreaState = OemState.invalid,
        dsid = [ 53, 26, 0, 0],
        serialNumber = 0x12345,
    )
    repr(DSDevice(**devSpec))

    devSpec = dict(
        groups = [1, 2, 3, 4],
        terminalBlocks = [tb],
        oemAreaState = OemState.valid,
        dsid = [ 53, 26, 0, 0],
        oemEan = 0x12345678ab,
        oemInetNeed = 0x1,
        oemSerialNumber = 0x123445,
    )
    repr(DSDevice(**devSpec))
