#!/usr/bin/env python

from ds485d_mock import *
from ds485c_mock import *
from notification_queue import *
from process_pkt import *
from ds485.dsuid import *
from util import list2hexstr
from iptables import *
from testsuite_harness import *

from Queue import Queue, Empty
import unittest
import time, datetime
import sys, os
import traceback, signal
import logging
import asyncore, socket
import random
from threading import Thread, Timer

TIMESTMP_FMT = ("%H:%M:%S") # no %f for msecs
FORMAT = '%(asctime)s.%(msecs)03d %(name)-5s %(message)s'
logging.basicConfig(format=FORMAT, datefmt=TIMESTMP_FMT, level=logging.INFO)

class Frack(unittest.TestCase):
    log = logging.getLogger("testsuite")

    def setUp(self):
        log = self.log
        #log.setLevel(logging.DEBUG)
        log.debug("setUp")
        log.debug("PID %d", os.getpid())
        signal.signal(signal.SIGINT, self.signal_handler)

        try:
            ports = [ 5004, 5005 ]
            self.ds485_port = 8445

            os.system("pkill -9 ds485p")
            for p in ports:
                iptables_unblock(log, p)

            fork_ds485p(self.ds485_port, ports,
                        loglevel = 5, valgrind = True)

            self.select_loop = AsyncCore()
            self.select_loop.start()

            # open the socket after forking
            self.daemons = fire_up_daemons(ports)
            if not self.daemons:
                log.info("")
                log.info("ds485p failed to connected to our fake-buses")
                log.info("")
                seppuku()


            log.info("")
            log.info("ds485p connected to our fake-buses")
            log.info("")

        except Exception, e:
            print traceback.format_exc()
            raise e # TODO is tearDown called?


    def signal_handler(self, signal, frame):
        self.log.warning('You pressed Ctrl+C!')
        seppuku()

    def tearDown(self):
        self.log.info("tearDown")
        kill_ds485p()
        asyncore.close_all()
        for d in self.daemons:
            d.close()
        self.select_loop.stop()

    def stop(self):
        self.log.info("do stop")
        self.run = False

    #@unittest.skip("")
    def test_dsm_reboots(self):
        log = self.log
        log.info("")
        log.info("DSM reboot tests")
        log.info("")
        log.info("Setup: 2 busses each having 2 DSM, 1 client listening for bus change")
        log.info("messages. Fetch device list from ds485p bus, then emit random(1,5)")
        log.info("bus change indicating leave/join of dsm through the busses. Check that")
        log.info("the client sees them all and only those we emitted.")
        log.info("Aftwards let the client refetch the complete device list (snapshot list)")
        log.info("and compare it 1) with the list compute out of previous snapshot plus")
        log.info("the incremental bus changes (incremental list) 2) comare the snapshot")
        log.info("with device list of the busses. Repeat this until 37 bus changes are")
        log.info("emitted.")
        log.info("")

        dsmids = []
        dsmids_to_src = {}
        for d in self.daemons:
            dsmids.extend(d.state.dsmids)
            for dsm in d.state.dsmids:
                dsmids_to_src[dsm] = d

        class states:
            initialize = 'i'
            send_new_bc = 's'
            wait_bc = 'w'
            fetch_devices = 'f'
            wait_fetch_devices = 'g'
            crash = 'c'

        tstate = states.initialize
        off = []
        received_bus_changes = []
        pending_msg = []
        old_dsuid_list = []
        ct = 0
        cutoff = 37 # max messages
        countdown = 0
        client = None

        self.run = True
        try:
            while self.run:
                try:
                    (evt, pkt, src) = [None] * 3
                    (evt, pkt, src) = notification_queue.get(True, 1)
                    reply = None
                except Empty: # timeout

                    countdown -= 1
                    if countdown > 0:
                        continue

                    if not client:
                        log.info("Start Client")
                        client = Client('localhost', self.ds485_port, ClientState())
                        continue

                    cstate = client.state

                    if tstate == states.initialize:
                        log.info("initialize")
                        if not cstate.connected:
                            client.push(bytearray(createQueryBusState(0)))
                            continue

                        # wait for initial bus changes from sysConnThread
                        countdown = 20
                        cstate.dsmids = list(dsmids) # hack, dsmids_incremental
                        old_dsuid_list = list(cstate.dsmids)
                        tstate = states.fetch_devices
                        continue

                    elif tstate == states.fetch_devices:
                        log.info("fetch devices")

                        dsmids_incremental = cstate.dsmids
                        cstate.dsmids = [ 'sentinel' ]
                        client.push(bytearray(createDeviceListQuery()))
                        tstate = states.wait_fetch_devices

                    elif tstate == states.wait_fetch_devices:
                        crash = False

                        log.info("wait fetch devices")
                        if cstate.dsmids == [ 'sentinel' ]:
                            continue

                        # verify snapshot matches incremental dsm join/leave
                        # updates
                        if ((len(dsmids_incremental) != len(cstate.dsmids)) or
                            (set(dsmids_incremental) != set(cstate.dsmids))):

                            log.info("device list snapshot/incremental mismatch")
                            cstate.dump_device_list();

                            log.info("device list incremental")
                            for dsm in dsmids_incremental:
                                log.info("\t%s", dsm)

                            log.info("bus changes received")
                            for (dsuid, join) in received_bus_changes:
                                log.info("%s %s", dsuid, "join" if join else "left")

                            log.info("device list before emitting bus changes")
                            for dsm in old_dsuid_list:
                                log.info("\t%s", dsm)

                            crash = True

                        # verify device list of client reflects deamons list
                        tmp = []
                        [ tmp.extend(s.state.dsmids) for s in self.daemons ]

                        if ((len(tmp) != len(cstate.dsmids)) or
                            (set(tmp) != set(cstate.dsmids))):

                            log.info("device list client/deamon mismatch")
                            cstate.dump_device_list();
                            for d in self.daemons:
                                d.dump_device_list()

                            crash = True

                        if crash:
                            countdown = 10
                            tstate = states.crash
                            continue

                        log.info("device lists match")
                        cstate.dump_device_list();

                        if ct < cutoff:
                            tstate = states.send_new_bc
                            continue

                        self.run = False
                        log.info("")
                        log.info("              SUCCESS")
                        log.info("")

                    elif tstate == states.send_new_bc:

                        log.info("send_new_bc")

                        old_dsuid_list = list(cstate.dsmids)
                        pending_msg = []
                        received_bus_changes = []
                        for i in range(random.randint(1, 5)):
                            ct += 1
                            cur = dsmids[random.randint(0, len(dsmids) - 1)]
                            if cur in off:
                                off.remove(cur)
                                m = createBusChange(cur.octets, 0)
                            else:
                                off.append(cur)
                                m = createBusChange(cur.octets, 1)

                            pending_msg.append((cur, cur not in off))
                            log.info("bus_change%d: %s dsuid: %s", ct,
                                     "leave" if (cur in off) else "join ", cur)

                            d = dsmids_to_src[cur]
                            if cur in off:
                                d.state.dsmids.remove(cur)
                            else:
                                d.state.dsmids.append(cur)
                            d.send_all_clients(bytearray(m))

                        countdown = len(pending_msg) * 4 + 2 # 3s sleep in ds485p
                        tstate = states.wait_bc
                        continue

                    elif tstate == states.wait_bc:
                        log.info('wait bus changes')

                        if ((len(received_bus_changes) != len(pending_msg)) or
                            (set(received_bus_changes) != set(pending_msg))):
                            log.info("bus changes mismatch")

                            log.info("pending bus changes:")
                            for i in pending_msg:
                                log.info("\t%r", i)

                            log.info("received bus changes:")
                            for i in received_bus_changes:
                                log.info("\t%r", i)

                            self.assertTrue(False)


                        log.info("bus changes match")
                        tstate = states.fetch_devices

                    elif tstate == states.crash:
                        log.info("")
                        log.info("              FAILURE")
                        log.info("")

                        self.run = False
                        self.assertTrue(False)

                    continue # otherwise we handle a stale pkt

                #log.debug("evt: %s pkt: %s src: %r", evt,
                #          list2hexstr(pkt) if pkt else None, src)

                if isinstance(src, Connection):

                    if (evt == handle_read):
                        reply = process_pkt(pkt, log, src.state)

                elif isinstance(src, Client):

                    if (evt == handle_connect):
                        log.info("handle_connect")
                        reply = createQueryBusState(0)
                    elif evt == handle_close:
                        log.info("server closed connection, reconnect")
                        client = None
                    elif (evt == handle_read):
                        # behave like a proper client
                        reply = process_pkt_client(pkt, log, src.state)

                        if pkt[0] == CMD.DS485D_CMD_BUS_CHANGE:
                            bc = parseBusChange(pkt)
                            received_bus_changes.append((bc.dsuid, bc.joined))


                if not reply:
                    continue

                #log.info("reply: %s",  list2hexstr(reply))
                src.push(bytearray(reply))
                reply = None

            if client:
                client.close()

        except KeyboardInterrupt, e:
            print traceback.format_exc()
            pass
        except AssertionError, e:
            raise
        except Exception, e:
            print traceback.format_exc()
            pass

        log.info("exiting...")

if __name__== "__main__":
    unittest.main()
    #seppuku()
